#! /bin/bash

#**************************************************************************
#    begin                : Wed July 11, 2001
#    copyright            : (C) 2001 by Adaptec
#    email                : deanna_bonds@adaptec.com
#
#    This is the installer script for Adaptec's dpt_i2o RAID driver
#**************************************************************************
#                                                                         *
#   This program is free software; you can redistribute it and/or modify  *
#   it under the terms of the GNU General Public License as published by  *
#   the Free Software Foundation; either version 2 of the License, or     *
#   (at your option) any later version.                                   *
#                                                                         *
#**************************************************************************

installer_path=/usr/adaptec/i2o_driver
PATH=$installer_path:$PATH 
export PATH 

VERSION="1.1"
verbose=0
root_dir=""
do_mkinitrd=1
do_lilo=1
force=0
module_count=0

echo
echo Adaptec I2O RAID driver installer - V$VERSION
echo 

lilo_conf=$root_dir"/etc/lilo.conf"
prebuilt_binaries=/usr/adaptec/i2o_driver/dpt_i2o_prebuilt.cgz
tmp_magic=/tmp/magic$$

# output usage statement
usage () {
	echo "usage: `basename $0` [--version] || [--help] ||" >&2
	echo "            [--verbose] [--root=<root dir>] [--noinitrd] [--nolilo] [--force]" >&2
	echo
	echo "usage: `basename $0` [--list_modules] " >&2
	echo
	echo "For the proceedure to patch your kernel sources see /usr/adaptec/i2o_driver/Read.me"
	echo
	exit 1
}

# write an updated magic file
write_magic () {
cat <<-blub >$tmp_magic
	0x202	string		HdrS		Linux/i386 Kernel
	>0x206	short		0x0105		setupV1 zImage
	>0x206	short		0x0200		setupV2
	>>0x211	byte		0		zImage
	>>0x211	byte		1		bzImage
	>(0x20e.s+0x205)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%5.5s"
	>(0x20e.s+0x206)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%6.6s"
	>(0x20e.s+0x207)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%7.7s"
	>(0x20e.s+0x208)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%8.8s"
	>(0x20e.s+0x209)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%9.9s"
	>(0x20e.s+0x20a)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%10.10s"
	>(0x20e.s+0x20b)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%11.11s"
	>(0x20e.s+0x20c)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%12.12s"
	>(0x20e.s+0x20d)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%13.13s"
	>(0x20e.s+0x20e)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%14.14s"
	>(0x20e.s+0x20f)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%15.15s"
	>(0x20e.s+0x210)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%16.16s"
blub
}

clean_magic () {
	rm -f $tmp_magic 2>/dev/null
}

list_prebuilt_modules () {

	echo Prebuilt modules:
	list=`zcat $prebuilt_binaries | cpio --list 2>/dev/null | sort` 
	display_list=`echo $list | sed "s/dpt_i2o\.//g"`
	for i in $display_list ; do
		echo "	** $i"
	done
	
	echo
}


install_module() {
	if ! [ -e $2 ] ; then
		mkdir -p $2 
	fi
	cd $2
	module_file=dpt_i2o.$1

	if [ $verbose -eq 1 ] ; then
		echo "Extracting "$module_file
	fi
	gzip -cd $prebuilt_binaries | cpio --extract dpt_i2o.$1 > /dev/null 2> /dev/null
	if ! [ -e dpt_i2o.$1 ] ; then
		echo "No prebuilt module for "$1
		return 1
	fi	   
	mv -f dpt_i2o.$1 dpt_i2o.o
	if [ $verbose -eq 1 ] ; then
		echo "Module installed for "$1" kernel."
	fi
	return 0
}


modify_rc_config() {
	rc_config=$root_dir"/etc/rc.config"
	if [ $verbose -eq 1 ] ; then
		echo "Modifying "$rc_config
	fi

	if ! [ -e $rc_config ] ; then
		if [ $verbose -eq 1 ] ; then
			echo "No rc.config file"
		fi
		return 1
	fi

        # Backup current config
        cp -fp $rc_config $root_dir/etc/rc.config.$$
	if [ $verbose -eq 1 ] ; then
		grep "INITRD_MODULES" $rc_config
	else
		grep "INITRD_MODULES" $rc_config >/dev/null
	fi
        if [ $? -ne 0 ] ; then
		# Configuration not set up to use ramdisks add it
		echo "INITRD_MODULES= \"\"" >>$rc_config 
        fi
	sed "s/INITRD_MODULES[ \t]*=[ \t]*/INITRD_MODULES=/" $rc_config >/tmp/rc.config.tmp && mv -f /tmp/rc.config.tmp $rc_config

        current_modules=`grep "INITRD_MODULES[ \t]*=" $rc_config | head -1`
        echo $current_modules | grep dpt_i2o >/dev/null
	if [ $? -ne 0 ] ; then
		# If we cannot find dpt_i2o in the config file, add it.
		new_modules=`echo $current_modules | sed "s/\"[ \t]*$/ dpt_i2o\"/"`
		sed "s/INITRD_MODULES.*/$new_modules/" $rc_config >/tmp/rc.config.tmp && mv -f /tmp/rc.config.tmp $rc_config
        fi
	if [ $verbose -eq 1 ] ; then
		grep "INITRD_MODULES" $rc_config
	fi
	return 0
}


modify_modules_conf() {

	modules_conf=""
	if [ -e $root_dir"/etc/modules.conf" ] ; then
		modules_conf="$root_dir/etc/modules.conf"
	elif [ -e $root_dir"/etc/conf.modules" ] ; then
		modules_conf=$root_dir"/etc/conf.modules"
	else
		if [ $verbose -eq 1 ] ; then
			echo "No modules.conf file"
		fi
		return 1
	fi
	if [ $verbose -eq 1 ] ; then
		echo "Modifying "$modules_conf
	fi

	grep dpt_i2o $modules_conf > /dev/null
	if [ $? -ne 0 ] ;  then
		# Put dpt_i2o entry if not already in config file
		echo "alias scsi_hostadapter dpt_i2o" >> $modules_conf
		if [ $verbose -eq 1 ] ; then
			echo "No dpt_i2o entry - adding it..."
		fi
	fi

	tmp_modules=/tmp/modules.conf.$$
	# Enumerate scsi_hostadapter entries
	awk ' \
	BEGIN \
	{ \
		c=1; \
	} \
	{ \
		if (($1 == "alias") && (substr($2, 0, length("scsi_hostadapter")) == "scsi_hostadapter")) \
		{ \
			$2="scsi_hostadapter"c; \
			print $0; \
			c=c+1; \
		} \
		else \
		{ \
			print $0; \
		} \
	}' $modules_conf >$tmp_modules && mv -f $tmp_modules  $modules_conf
	if [ $verbose -eq 1 ] ; then
		grep dpt_i2o $modules_conf 
	fi

}

run_mkinitrd() {

        echo "Creating initrd..."
# $1 = image
# $2 = kernel version
# $3 = initrd

	old_initrd=$3.$$
	mv -f $3 $old_initrd 2>/dev/null >/dev/null
	mkinitrd=`which mk_initrd 2>/dev/null`

	if [ -z "$mkinitrd" ] ; then	
		# if not the SuSE mk_initrd then use Red Hat style
		if [ "$root_dir" != "" ] ; then
			# if we specified root then use our modified one
			mkinitrd="/usr/adaptec/i2o_driver/adpt_mkinitrd --root="$root_dir
		else
			mkinitrd=`which mkinitrd 2>/dev/null`
		fi

		if [ -z "$mkinitrd" ] ; then
			echo "Can not find mkinitrd. Exitting..."
			exit 1
		fi

		if [ $verbose == 1 ] ; then
			echo "$mkinitrd $3 $2"
		fi

		$mkinitrd $3 $2
	else
		grep root_device $mkinitrd >/dev/null 2>/dev/null
		if [ $? -ne 0 ] ; then
			mkinitrd="/usr/adaptec/i2o_driver/adpt_mk_initrd"
		fi

		if [ "$root_dir" != "" ] ; then
			mkinitrd=$mkinitrd" "$root_dir
		fi
		if [ $verbose == 1 ] ; then
			echo "$mkinitrd -k `basename $1` -i `basename $3`"
		fi

		$mkinitrd  -k `basename $1` -i `basename $3`
	fi

	if [ $? -ne 0 ] ; then
		echo Problem making initrd.  Quitting.
		mv -f $old_initrd $3
		return 1
	fi
	return 0

}

run_lilo() {

	#TODO add root dir options
	echo
        echo "Running lilo..."
	if [ "$root_dir" != "" ] ; then
		lilo -r $root_dir -w
	else
		lilo
	fi
	if [ $? -ne 0 ] ; then
		echo Problem running lilo.  Quitting.
		exit 1
	fi
	echo
        echo "Reboot to load Ram Disk."
	echo
}



create_device_nodes() {
	if [ $verbose -eq 1 ] ; then
		echo "Creating Device Nodes..."
	fi
	num=0
	while [ $num -lt 4 ]
	do
	    if [ ! -c $root_dir/dev/dpti$num ]
	    then
		mknod $root_dir/dev/dpti$num c 151 $num
	    fi
	    num=`expr $num + 1`
	done
}




# parse the arguments
while [ $# -gt 0 ]; do
	case $1 in
		--version)
			echo `basename $0`": version $VERSION"
			exit 0
			;;

		--verbose)
			verbose=1
			;;

		--root*)
			if echo $1 | grep '=' >/dev/null ; then
				root_dir=`echo $1 | sed 's/^--root=//'`
			else
				root_dir=$2
				shift
			fi
			;;

		--list_modules)
			list_prebuilt_modules
			exit 0
			;;

		--noinitrd)
			do_mkinitrd=0
			do_lilo=0
			;;

		--nolilo)
			do_lilo=0
			;;

		--force)
			force=1
			;;

		--help)
			usage	
			;;
		*)
			usage
			;;
	esac

	shift
done

if [ $verbose -eq 1 ] ; then
	echo "Options: do_mkinitrd="$do_mkinitrd" do_lilo="$do_lilo" root dir="$root_dir
fi

#for now if no lilo then we can not install anything. 
# TODO next version install anyway to the /lib/modules directories
if ! [ -e $lilo_conf ] ; then
	echo "System is not configured to use lilo.  Can not install."
	exit 1
fi

modify_rc_config
modify_modules_conf
write_magic 

file=file
[ -x /usr/bin/file ] && file=/usr/bin/file
[ -x $root_dir/usr/bin/file ] && file=$root_dir/usr/bin/file

awk '
BEGIN {
	FS="[ \11=]*"
	global=1
	global_initrd="?"
	global_label="?"
	image=""
	initrd="?"
	label="?"
	isImage=0
}
/image[ \11]*=/ {
	if ( global == 0 && isImage == 1) {
		print image " " label " " initrd
		label="?"
		initrd="?"
	}
	isImage = 1
	if ( $1 == "" ) {
		image=$3
	} else {
		image=$2
	}
	label="?"
	initrd="?"
	if (( global_initrd != "?" ) && ( global != 0 )) {
		print "? " global_label " " global_initrd
	}
	global=0
}
/label[ \11]*=/ {
	if ( $1 == "" ) {
		label=$3
	} else {
		label=$2
	}
	if ( global != 0 ) {
		global_label=$2
	}
}
/initrd[ \11]*=/ {
	if ( $1 == "" ) {
		initrd=$3
	} else {
		initrd=$2
	}
	if ( global != 0 ) {
		global_initrd=$2
	}
}
/other[ \11]*=/ {
	if ( global == 0 && isImage == 1 ) {
		print image " " label " " initrd
		label="?"
		initrd="?"
	}
	isImage = 0
	label="?"
	initrd="?"
	image="?"
	if (( global_initrd != "?" ) && ( global != 0 )) {
		print "? " global_label " " global_initrd
	}
	global=0
}
END {
	if ( global == 0 && isImage == 1) {
		print image " " label " " initrd
	}
}' $lilo_conf | sort +2 | while read image label initrd; do

	if [ $image == "?" ] ; then
		global_initrd=$initrd
		continue
	fi
	echo
	file_magic=`$file -L -m $tmp_magic $image`" \""
	kernel=`echo $file_magic | cut -d \" -f 2 `

	if [ -z $kernel ] ; then
		echo
		echo ***NOTICE: $image " is not a kernel. Skipping."
		continue
	fi

	if [ $verbose -eq 1 ] ; then
		echo "label="${label}"	image="${image}"	version="$kernel"	initrd="${initrd}
	fi

	if [ $do_mkinitrd == 1 ] ; then
		if [ $initrd == "?" ] ; then
			if [ -z $global_initrd ] ; then
				echo
				echo "****** PROBLEM *******"
				echo Kernel $image is not configured to use initrd.  Skipping.
				echo "Modify /etc/lilo.conf for this config to use an initrd"
				echo "(e.g. initrd=/boot/initrd-$kernel.img)"
				echo "**********************"
				continue
			fi
			if [ -z $last_initrd_kernel ] ; then
				initrd=$global_initrd
			elif [ $last_initrd_kernel == $kernel ] ; then
				initrd=$global_initrd
			else
				echo "****** PROBLEM *******"
				echo A globally declared initrd is used for two different 
				echo kernel versions.  Please fix this in /etc/lilo.conf
				echo and rerun the install.
				echo "**********************"
				exit 1
			fi
			if [ $verbose -eq 1 ] ; then
				echo "Setting initrd to global initrd->"$initrd
			fi
			last_initrd_kernel=$kernel

		fi

		if [ x$initrd == x$last_initrd ] ; then
			if [ x$kernel == x$last_kernel ] ; then
				echo "Same as last initrd. Skipping."
				continue;
			elif [ $force -ne 1 ] ; then
				echo "****** PROBLEM *******"
				echo In lilo.conf there are two different kernel versions 
				echo assigned to the same initrd - fix this and rerun installer.
				echo Or use the --force option
				echo "**********************"
				exit 1
			fi
		fi
	fi


	echo $image" -> "$kernel
	last_kernel=$kernel
	last_initrd=$initrd
	tmpVersion=`echo $kernel | sed -e "s/^2.2//"`
	if [ $tmpVersion != $kernel ] ; then
		installPath=$root_dir"/lib/modules/"$kernel"/scsi"
	else
		installPath=$root_dir"/lib/modules/"$kernel"/kernel/drivers/scsi"
	fi
	echo "Installing in "$installPath
	
	install_module $kernel $installPath
	if [ $? -ne 0 ] ; then
		continue;
	fi

	module_count=`expr $module_count + 1`
	echo $module_count >/tmp/dpti.tmp

	if [ $do_mkinitrd == 1 ] ; then
		run_mkinitrd $image $kernel $initrd
		if [ $? -ne 0 ] ; then
			exit 1
		fi
	fi

done

if [ $? -eq 1 ] ; then
	clean_magic
	exit 1
fi

clean_magic
module_count=`cat /tmp/dpti.tmp`
rm -f /tmp/dpti.tmp >/dev/null 2>/dev/null
if [ $module_count -eq 0 ] ; then
	echo
	echo "***************************************"
	echo "******** NO MODULES INSTALLED *********"
	echo "***************************************"
	echo
	echo "Done."
	exit 1
fi


echo

if [ $do_lilo == 1 ] ; then
	run_lilo	
fi

create_device_nodes

echo Done.
exit

